#include <msp/core/maputils.h>
#include <msp/datafile/collection.h>
#include "bindable.h"
#include "gl.h"
#include "font.h"
#include "immediate.h"
#include "primitivetype.h"
#include "texture2d.h"

using namespace std;

namespace Msp {
namespace GL {

Font::Font():
	native_size(1),
	ascent(1),
	descent(0),
	cap_height(1),
	x_height(0.5)
{ }

// Avoid synthesizing ~RefPtr in files including font.h
Font::~Font()
{ }

void Font::set_texture(const Texture2D &t)
{
	texture = &t;
	texture.keep();
}

const Texture2D &Font::get_texture() const
{
	if(!texture)
		throw logic_error("No texture");
	return *texture;
}

void Font::add_glyph(const Glyph &g)
{
	insert_unique(glyphs, g.code, g);
}

void Font::set_kerning(unsigned l, unsigned r, float d)
{
	kerning[KerningKey(l, r)] = d;
}

float Font::get_string_width(const string &str, StringCodec::Decoder &dec) const
{
	float x = 0;

	unsigned prev = 0;
	for(string::const_iterator i=str.begin(); i!=str.end();)
	{
		unsigned c = dec.decode_char(str, i);
		if(prev)
			x += get_glyph_advance(prev, c);
		prev = c;
	}
	x += get_glyph_advance(prev);

	return x;
}

void Font::draw_string(const string &str, StringCodec::Decoder &dec, const Color &color) const
{
	BindRestore bind_tex(get_texture());
	Immediate imm((TEXCOORD2, COLOR4_UBYTE, VERTEX2));
	imm.color(color);
	build_string(str, dec, imm);
}

void Font::build_string(const string &str, StringCodec::Decoder &dec, PrimitiveBuilder &bld) const
{
	MatrixStack::Push push_mtx(bld.matrix());

	unsigned prev = 0;
	for(string::const_iterator i=str.begin(); i!=str.end();)
	{
		unsigned c = dec.decode_char(str, i);
		GlyphMap::const_iterator j = glyphs.find(c);
		if(j==glyphs.end())
			continue;

		if(prev)
			bld.matrix() *= Matrix::translation(get_glyph_advance(prev, c), 0, 0);

		create_glyph_quad(j->second, bld);
		prev = c;
	}
}

void Font::create_glyph_quad(const Glyph &glyph, PrimitiveBuilder &bld) const
{
	bld.begin(TRIANGLE_STRIP);
	bld.texcoord(glyph.x1, glyph.y2);
	bld.vertex(glyph.off_x, glyph.off_y+glyph.h);
	bld.texcoord(glyph.x1, glyph.y1);
	bld.vertex(glyph.off_x, glyph.off_y);
	bld.texcoord(glyph.x2, glyph.y2);
	bld.vertex(glyph.off_x+glyph.w, glyph.off_y+glyph.h);
	bld.texcoord(glyph.x2, glyph.y1);
	bld.vertex(glyph.off_x+glyph.w, glyph.off_y);
	bld.end();
}

float Font::get_glyph_advance(unsigned code, unsigned next) const
{
	GlyphMap::const_iterator i = glyphs.find(code);
	if(i==glyphs.end())
		return 0;

	float advance = i->second.advance;

	if(next)
	{
		KerningMap::const_iterator j = kerning.find(KerningKey(code, next));
		if(j!=kerning.end())
			advance += j->second;
	}

	return advance;
}


Font::Loader::Loader(Font &f):
	DataFile::CollectionObjectLoader<Font>(f, 0)
{
	init();
}

Font::Loader::Loader(Font &f, Collection &c):
	DataFile::CollectionObjectLoader<Font>(f, &c)
{
	init();
}

void Font::Loader::init()
{
	add("native_size", &Font::native_size);
	add("ascent",      &Font::ascent);
	add("cap_height",  &Font::cap_height);
	add("descent",     &Font::descent);
	add("texture",     &Loader::texture);
	add("texture",     &Loader::texture_ref);
	add("glyph",       &Loader::glyph);
	add("kerning",     &Loader::kerning);
	add("x_height",    &Font::x_height);
}

void Font::Loader::glyph(unsigned c)
{
	Glyph gl;
	gl.code = c;
	load_sub(gl);
	obj.glyphs.insert(GlyphMap::value_type(c, gl));
}

void Font::Loader::kerning(unsigned l, unsigned r, float d)
{
	obj.kerning[KerningKey(l, r)] = d;
}

void Font::Loader::texture()
{
	RefPtr<Texture2D> tex = new Texture2D;
	load_sub(*tex);
	obj.texture = tex;
}

void Font::Loader::texture_ref(const string &name)
{
	obj.texture = &get_collection().get<Texture2D>(name);
	obj.texture.keep();
}


Font::Glyph::Loader::Loader(Glyph &g):
	DataFile::ObjectLoader<Glyph>(g)
{
	add("texcoords", &Loader::texcoords);
	add("size",      &Glyph::w,     &Glyph::h);
	add("offset",    &Glyph::off_x, &Glyph::off_y);
	add("advance",   &Glyph::advance);
}

void Font::Glyph::Loader::texcoords(float x1_, float y1_, float x2_, float y2_)
{
	obj.x1 = x1_;
	obj.y1 = y1_;
	obj.x2 = x2_;
	obj.y2 = y2_;
}

} // namespace GL
} // namespace Msp
